//=================== Klassen - Implementierung: PERSPEKTIVE ============================
// Perspektive-Klasse fuer 3D-Zentral- oder Parallel-Perspektive mit rot/grner Brille
// Matrizen werden spaltenweise gespeichert !!!
// ======================================================================================
// Urheberrecht: Univ.-Prof. Dr.-Ing. Udo F. Meissner, Darmstadt
// Weiterfuehrendes Lehr- und Studienprogramm - kommerzielle Nutzung nicht gestattet 
// Der Autor bernimmt keine Haftung fr Unvollstndigkeiten und Fehler der Algorithmen!
// Version: 1994 / 2023
// Compiler: C++ 
// ================ Include-Files =======================================================
#include "framework.h"
#include "PERSPEKTclass.h"
//
// #include "Perspect.h"
// #include "MATRIX.h"
// #include "VEKTOR.h"
// ============================================================================
// public -> Oeffentliche Mitgliedsfunktionen
// ==========================================
//
// Konstruktor -> PERSPEKTIVE per(A, F, H, art, augen);
// ====================================================
PERSPEKTIVE::PERSPEKTIVE(VEKTOR &A,VEKTOR &F,VEKTOR &H, int art,int augen) : persp_art(art), anz_augen(augen)
{
//*** std::cout<<"PERSPEKTIVE::Konstruktor: neue Instanz von PERSPEKTIVE erzeugen "<<'\n'; 
//-------------------------------------------------------------------------
// Augenabstand [cm]
   augen_abstand=6.75;
// ###################
// Augenpunkt A der Welt
   A_welt=new VEKTOR(3);
// Horizontpunkt
   H_welt=new VEKTOR(3);
// Vektor z(3) vom Augenpunkt A zum Referenzpunkt O 
   z=new VEKTOR(3);
// Transformationsmatrix g(3,3)
   g=new MATRIX(3,3);
//----------------------------------------------------------------------
// Zwischenspeicher
// Bezugsachsen
VEKTOR g1(3),g2(3),g3(3);
// e-Matrix
   e_MATRIX e(3,3,3);
//***cout<<"e: "<<e;
//----------------------------------------------------------------------
// Augenpunkt speichern
   *A_welt=A;
// ---------------------------------------------------------------------
// Horizontpunkt speichern
   *H_welt = H; 
//----------------------------------------------------------------------
// Bezugsachse g3(x,y,z) ermitteln
// -------------------------------
   g3=A-F;
   g3=g3.normieren();
//***cout<<"Bezugsachse g3(x,y,z): "<<g3;
//
// Bezugsachse g1(x,y,z), z_abst_O, h_entf_horizont_O ermitteln
// ---------------------------------------------------------------
   *z=A-H;
   z_abst_O=*z*g3;
//*** cout<<"Abstand des Augenpunktes A vom Referenzpunkt O der Projektionswand: "<<z_abst_O<<'\n';
   *z=-z_abst_O*g3;
// maximaler Abstand des Auge zum Objektpunkt x
   max_abst_X = 0.0;
// Basisvektor g1 
   g1=H-A-*z;
   h_entf_horizont_O=g1.betrag();
   g1=g1/h_entf_horizont_O;
//***cout<<"Bezugsachse g1(x,y,z): "<<g1;
//***cout<<"Abstand des Horizontpunktes H vom Referenzpunkt O= "<<h_entf_horizont_O<<'\n';
//
// Bezugsachse g2(x,y,z) = g3 x g1 = e(i,j,k) * g3(j) * g1(k) ermitteln
// ------------------------------------------------------------------------
   g2=(e*g3)*g1;
//***cout<<"Bezugsachse g2(x,y,z): "<<g2;
//----------------------------------------------------------------------
// Transformationsmatrix g=[g1 g2 g3] aufbauen
   *g=g1;
   g->teilmatrix_hinein(0,1)=g2;
   g->teilmatrix_hinein(0,2)=g3;
//***cout<<"Transformationsmatrix g: "<<*g;
//----------------------------------------------------------------------
// minimale und maximale Bildkoordinaten vorbelegen
   min1=min2=max1=max2=0.0;
};//--------------------------------------------------------------------
//
// Destruktor -> delete &per;
// ==========================
PERSPEKTIVE::~PERSPEKTIVE(void)
{
//*** std::cout<<"PERSPEKTIVE::Destruktor: Instanz von PERSPEKTIVE loeschen "<<'\n';
//          ---------------------------------------------------------
// ----------------------------------------------------- 
// maximaler Abstand Objektpunkt x zur Projektionsflche
// bei Stereo ggfls. Projektionsflche verlegen
    if(anz_augen > 1)
    {
        double diff_abst = max_abst_X - z_abst_O;
        // diff_abst + 1 promille wegen Aufrundung von H
        diff_abst = diff_abst * 1.001;                         
        if (diff_abst > 0)
        {
            cout << endl;
            cout << "Korrektur: Die Projektionsflaeche (Horizontpunkt H) muss hinter die Objektpunkte X verlegt werden!!!"
                 << endl;
            *H_welt = *H_welt + (*z) * diff_abst / z->betrag();          
            cout << "Vorschlag: H = " << *H_welt;
        };
    };
// ----------------------------------------------------
// Matrizen loeschen
   delete A_welt;
   delete H_welt;
   delete z; 
   delete g;
};//--------------------------------------------------------------------
//
// Augenpunkt -> a_bild=per.auge(&z_abstand_O);
// =============================================
MATRIX& PERSPEKTIVE::auge(double &z_abstand_O)
{
// Augenpunkt a(2) des Bildes
VEKTOR& a_bild = *new VEKTOR(2);
// Abstand des Augenpunktes A vom Referenzpunkt O
z_abstand_O=z_abst_O;
// Bildkoordinaten des Augenpunktes A
   a_bild(0) = 0.0;
   a_bild(1) = 0.0;
return a_bild;
};//--------------------------------------------------------------------
//
// Horizontpunkt -> h_bild=per.horizont();
// =======================================
MATRIX& PERSPEKTIVE::horizont()
{
// Horizontpunkt h(2) des Bildes
VEKTOR& h_bild = *new VEKTOR(2);
// Bildkoordinaten des Horizontpunktes H
   h_bild(0)=h_entf_horizont_O;
   h_bild(1)=0.0;
return h_bild;
};//--------------------------------------------------------------------
//
// Bildpunkt -> x_bild=per.transformation(X_welt,abst_entf_X,punkte);
// =======================================================================================
MATRIX& PERSPEKTIVE::transformation(MATRIX &X,MATRIX &abst_entf_X,int punkte)
{
// Abfrage, ob Entfernung der Objektpunkte gespeichter werden kann
// Matrixdimensionen von abst_entf_X ermitteln
   int nj, dummy;
   abst_entf_X.dimension_ermitteln(dummy, nj, dummy, dummy);
// 
// Bildpunkt
MATRIX& x_bild = *new MATRIX(2, anz_augen, punkte);
// Zwischenspeicher
   double abstand,alpha,wert,augab,entfernung;
VEKTOR E(3),Z(3),C(3);
// Bezugsachsen
VEKTOR g1(3),g3(3);
   g1=*g;
   g3=g->teilmatrix_heraus(0,2,3);
//----------------------------------------------------------------------
// E(x,y,z), abstand, Z(x,y,z), C(x,y,z) ermitteln
// -----------------------------------------------
// Schleife ueber die Punkte
// =========================
for(int pkt=0; pkt<punkte; pkt++)
{
 // Schleife ueber die Augen
 // ========================
   augab=0.0;
   if(anz_augen == 2) augab=augen_abstand*0.5;
   for(int auge=0; auge<anz_augen; auge++)
 {  
   // -----------------------------------------------
   E=X.teilmatrix_heraus(0,pkt,3) - (*A_welt-g1*augab);     // linkes Auge
   //*** cout<<"X"<<X;  cout<<"A"<<*A_welt;  cout<<"g1"<<g1; cout<<"augab"<<augab;
   entfernung = E.betrag();
   //*** cout<<"Entfernung des Objektpunktes X vom Augenpunkt A = "<<entfernung<<'\n';
   abstand=-g3*E;
   // ----------------------------------------------------------------------------------------------
   // maximalen Abstand vom Auge zum Objektpunkt x speichern
   // bei Stereo ggfls. Projektionsflche verlegen
   if (auge == 0)
     {
       if (abstand > max_abst_X) max_abst_X = abstand;
     }
   else
    {
       if (abstand > z_abst_O)
           cout << "Warnung: Projektionsflaeche liegt vor dem Objektpunkt X = " << X.teilmatrix_heraus(0, pkt, 3);
    };
   // -----------------------------------------------------------------------------------------------
   //*** cout<<"Abstand des Objektpunktes X vom Augenpunkt A = "<<abstand<<'\n';
   Z=-abstand*g3; 
// Bildpunkte der Parallelperspektive
   C=E-Z;  
//*** cout<<"x-y-z-Weltkoordinaten des Bildpunktes C "<<C;
  //----------------------------------------------------------------------
  // Transformation von C in Bildkoordinaten x(0,i,0) = g(0,s,i) * C(0,s) 
  // --------------------------------------------------------------------
    x_bild.teilmatrix_hinein(0,auge,pkt)=g->dimension(1,3,2)*C;
  //*** cout<<"Bildkoordinaten x_bild(1,2) in wahrer Groesse im 1-2-Bezugssystem:"<<x_bild;
  //----------------------------------------------------------------------
  // minimale und maximale Bildkoordinaten in wahrer Groesse (!) ermitteln
   wert=x_bild(0,auge,pkt);
   if(wert<min1) min1=wert;
   if(wert>max1) max1=wert;
//*** cout<<"min1= "<<min1<<" max1= "<<max1<<'\n';
  //
   wert=x_bild(1,auge,pkt);
   if(wert<min2) min2=wert;
   if(wert>max2) max2=wert;
//*** cout<<"min2= "<<min2<<" max2= "<<max2<<'\n';
  //----------------------------------------------------------------------
   if(persp_art>=1)
   {
    // Zentralperspektive
    alpha=*z*Z; 
    alpha=z_abst_O*z_abst_O/alpha;
//*** cout<<"alpha"<<alpha<<'\n';
    x_bild(0,auge,pkt)=x_bild(0,auge,pkt)*alpha - augab;
    x_bild(1,auge,pkt)=x_bild(1,auge,pkt)*alpha;
//*** cout<<"Bildkoordinaten x_bild(1,2) im 1-2-Bezugssystem:"<<x_bild;
   };
   if (anz_augen == 2) augab = augab-augen_abstand;        // rechtes Auge
 }; // Ende der Schleife ueber die Augen
//
// Ergebnisrueckgabe
           abst_entf_X(pkt,0) = abstand;
if(nj > 1) abst_entf_X(pkt,1) = entfernung;
//
};  // Ende der Schleife ueber die Punkte
//----------------------------------------------------------------------
// Ergebnisrueckgabe
return x_bild;
};//--------------------------------------------------------------------
//
// min./max. Bildfenster-Koordinaten -> per.fenster(min1,min2,max1,max2);
// ======================================================================
void PERSPEKTIVE::fenster(double &fmin1,double &fmin2,
                          double &fmax1,double &fmax2)
{
    fmin1=min1;
    fmin2=min2;
    fmax1=max1;
    fmax2=max2;
};//--------------------------------------------------------------------
